/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2009-2014 Baker and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// console.c

#include "environment.h"
#include "quakedef.h"

#ifndef _MSC_VER
#include <unistd.h>
#endif

console_t console1;


cbool	con_debuglog = false;


/*
================
Con_Quakebar -- johnfitz -- returns a bar of the desired length, but never wider than the console

includes a newline, unless len >= console1.buffer_columns.
================
*/
const char *Con_Quakebar (int len)
{
	static char bar[42];
	int i;

	len = c_min(len, (int)sizeof(bar) - 2);
	len = c_min(len, console1.buffer_columns);

	bar[0] = '\35';
	for (i = 1; i < len - 1; i++)
		bar[i] = '\36';
	bar[len-1] = '\37';

	if (len < console1.buffer_columns)
	{
		bar[len] = '\n';
		bar[len+1] = 0;
	}
	else
		bar[len] = 0;

	return bar;
}

/*
================
Con_ToggleConsole_f
================
*/
extern int history_line; //johnfitz

// Exiting the console ... set things.
void Con_Exit (void)
{
	History_Save (); // Baker: Keep history up-to-date.
	key_lines[edit_line][1] = 0;	// clear any typing
	Key_Console_Cursor_Move (0, cursor_reset); // Reset selection
	//key_linepos = 1; key_selstart = 0; key_sellength = 0;
	console1.backscroll = 0; //johnfitz -- toggleconsole should return you to the bottom of the scrollback
	history_line = edit_line; //johnfitz -- it should also return you to the bottom of the command history
}

void Con_ToggleConsole_f (lparse_t *unused)
{
	if (key_dest == key_console/* || (key_dest == key_game && console1.forcedup)*/)
	{
		Con_Exit ();
		if (cls.state == ca_connected)
		{
			Key_SetDest (key_game);
		}
		else
		{
			M_Menu_Main_f (NULL);
		}
	}
	else
	{
		Key_SetDest (key_console);
	}

	SCR_EndLoadingPlaque ();
	memset (console1.con_times, 0, sizeof(console1.con_times));
}

/*
================
Con_Clear_f
================
*/
void Con_Clear_f (void)
{
	if (console1.text)
		memset (console1.text, ' ', console1.buffer_size); //johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.backscroll = 0; //johnfitz -- if console is empty, being scrolled up is confusing
}

/*
================
Con_Dump_f -- johnfitz -- adapted from quake2 source
================
*/
void Con_Dump_f (void)
{
	int		l, x;
	const char	*line;
	FILE	*f;
	char	buffer[1024];
	char	name[MAX_OSPATH];

#if 1
	//johnfitz -- there is a security risk in writing files with an arbitrary filename. so,
	//until stuffcmd is crippled to alleviate this risk, just force the default filename.
	FS_FullPath_From_QPath (name, "condump.txt");
#else
	if (line->count > 2)
	{
		Con_Printf ("usage: condump <filename>\n");
		return;
	}

	if (line->count > 1)
	{
		if (strstr(line->args[1], ".."))
		{
			Con_Printf ("Relative pathnames are not allowed.\n");
			return;
		}
		FS_FullPath_From_QPath (name, line->args[1]);
		File_URL_Edit_Force_Extension (name, ".txt", sizeof(name));
	}
	else
		FS_FullPath_From_QPath (name, "condump.txt");
#endif

	f = FS_fopen_write_create_path (name, "w");
	if (!f)
	{
		Con_Printf ("ERROR: couldn't open file %s.\n", name);
		return;
	}

	// skip initial empty lines
	for (l = console1.cursor_row - console1.buffer_rows + 1 ; l <= console1.cursor_row ; l++)
	{
		line = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		for (x=0 ; x<console1.buffer_columns ; x++)
			if (line[x] != ' ')
				break;
		if (x != console1.buffer_columns)
			break;
	}

	// write the remaining lines
	buffer[console1.buffer_columns] = 0;
	for ( ; l <= console1.cursor_row ; l++)
	{
		line = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		strncpy (buffer, line, console1.buffer_columns);
		for (x = console1.buffer_columns - 1 ; x >= 0 ; x--)
		{
			if (buffer[x] == ' ')
				buffer[x] = 0;
			else
				break;
		}
		for (x=0; buffer[x]; x++)
			buffer[x] &= 0x7f;

		fprintf (f, "%s\n", buffer);
	}

	FS_fclose (f);
	Recent_File_Set_QPath (name);
	Con_Printf ("Dumped console text to %s.\n", name);
}

/*
================
Con_Copy_f -- Baker -- adapted from Con_Dump
================
*/
void Con_Copy_f (lparse_t *line)
{
	char	outstring[CONSOLE_TEXTSIZE]="";
	int		l, x;
//	char	*line;
	char	buffer[1024];

	if (line->count > 1) // More than just the command
	{
#ifdef SUPPORTS_WINDOWS_MENUS // Baker change
	// To be able to put this info on clipboard in menu
		if (strcasecmp (line->args[1], "vp80") == 0)
		{
			System_CopyToClipboard("http://www.optimasc.com/products/vp8vfw/index.html");
			Con_Printf ("Copied Google VP80/WebM Codec URL to clipboard.\n");
			return;
		}
#endif // Baker change +

		if (strcasecmp (line->args[1], "ents") == 0)
		{
			extern char *entity_string;
			if (!sv.active || !entity_string)
			{
				Con_Printf ("copy ents: Not running a server or no entities");
				return;
			}

			Clipboard_Set_Text (entity_string);

			Con_Printf ("Entities copied to the clipboard (%i bytes)\n", (int)strlen(entity_string));
			return;
		}

		if (strcasecmp (line->args[1], "screen") == 0)
		{
			SCR_ScreenShot_Clipboard_f ();
			return;
		}

		// Invalid args
		Con_Printf ("Usage: %s [ents | screen]: copies console to clipboard\n\"copy ents\" copies map entities to clipboard.\n\"copy screen\" copies screen to clipboard.", line->args[0]);
		return;
	}

	// skip initial empty lines
	for (l = console1.cursor_row - console1.buffer_rows + 1 ; l <= console1.cursor_row ; l++)
	{
		char	*myline  = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		for (x = 0 ; x < console1.buffer_columns ; x++)
			if (myline[x] != ' ')
				break;
		if (x != console1.buffer_columns)
			break;
	}

	// write the remaining lines
	buffer[console1.buffer_columns] = 0;
	for ( ; l <= console1.cursor_row ; l++)
	{
		char *myline = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		strncpy (buffer, myline, console1.buffer_columns);
		for (x = console1.buffer_columns - 1 ; x >= 0 ; x--)
		{
			if (buffer[x] == ' ')
				buffer[x] = 0;
			else
				break;
		}
		for (x=0; buffer[x]; x++)
			buffer[x] &= 0x7f;

		c_strlcat (outstring, va("%s\r\n", buffer));

	}

	Clipboard_Set_Text (outstring);
	Con_Printf ("Copied console to clipboard\n");
}

/*
================
Con_ClearNotify
================
*/
void Con_ClearNotify (void)
{
//	for (i = 0 ; i < scr_con_notifylines.value ; i++)
//		console1.con_times[i] = 0;

	memset (console1.con_times, 0, sizeof(console1.con_times));
}


/*
================
Con_MessageMode_f
================
*/
void Con_MessageMode_f (void)
{
	if (cls.state != ca_connected || cls.demoplayback)
		return;
	chat_team = false;
	Key_SetDest (key_message);
}

/*
================
Con_MessageMode2_f
================
*/
void Con_MessageMode2_f (void)
{
	if (cls.state != ca_connected || cls.demoplayback)
		return;
	chat_team = true;
	Key_SetDest (key_message);
}


/*
================
Con_CheckResize

If the line width has changed, reformat the buffer.
================
*/
void Con_CheckResize (void)
{
	int		i, j, width, oldwidth, oldtotallines, numlines, numchars;
	char	*tbuf; //johnfitz -- tbuf no longer a static array
	int mark; //johnfitz

	width = (vid.conwidth >> 3) - 2; //johnfitz -- use vid.conwidth instead of vid.width

	if (width == console1.buffer_columns)
		return;

	oldwidth = console1.buffer_columns;
	console1.buffer_columns = width;
	oldtotallines = console1.buffer_rows;
	console1.buffer_rows = console1.buffer_size / console1.buffer_columns; //johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	numlines = oldtotallines;

	if (console1.buffer_rows < numlines)
		numlines = console1.buffer_rows;

	numchars = oldwidth;

	if (console1.buffer_columns < numchars)
		numchars = console1.buffer_columns;

	mark = Hunk_LowMark (); //johnfitz
	tbuf = (char *) Hunk_Alloc (console1.buffer_size); //johnfitz

	memcpy (tbuf, console1.text, console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	memset (console1.text, ' ', console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE

	for (i=0 ; i<numlines ; i++)
	{
		for (j=0 ; j<numchars ; j++)
		{
			console1.text[(console1.buffer_rows - 1 - i) * console1.buffer_columns + j] =
					tbuf[((console1.cursor_row - i + oldtotallines) % oldtotallines) * oldwidth + j];
		}
	}

	Hunk_FreeToLowMark (mark); //johnfitz

	Con_ClearNotify ();

	console1.backscroll = 0;
	console1.cursor_row = console1.buffer_rows - 1;
}


/*
================
Con_Init
================
*/
void Con_Init (void)
{
	int i;
	char	temp[MAX_OSPATH]; // Was SYSTEM_STRING_SIZE
	const char	*t2 = "qconsole.log";

#pragma message ("Baker: OH fucking groovy.  No -condebug for dedicated since doesn't reach here.")
	con_debuglog = COM_CheckParm("-condebug");

	if (con_debuglog)
	{
		FS_FullPath_From_QPath (temp, t2);
		File_Delete (temp);
	}

	if (isDedicated)
		return;

	//johnfitz -- user settable console buffer size
	i = COM_CheckParm("-consize");
	if (i && i < com_argc-1)
		console1.buffer_size = c_max(CONSOLE_MINSIZE, atoi(com_argv[i+1]) * 1024);
	else
		console1.buffer_size = CONSOLE_TEXTSIZE;
	//johnfitz

	console1.text = (char *) Hunk_AllocName (console1.buffer_size, "context");//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	memset (console1.text, ' ', console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.buffer_columns = -1;

	//johnfitz -- no need to run Con_CheckResize here
	console1.buffer_columns = 78;
	console1.buffer_rows = console1.buffer_size / console1.buffer_columns;//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.backscroll = 0;
	console1.cursor_row = console1.buffer_rows - 1;
	console1.user_pct = 0.50; // Default console is 50%
	//johnfitz

	Con_SafePrintf ("Console initialized.\n");


	Cmd_AddCommands (Con_Init);

	console1.initialized = true;
}


/*
===============
Con_Linefeed
===============
*/
static void Con_Linefeed (void)
{
	//johnfitz -- improved scrolling
	if (console1.backscroll)
		console1.backscroll++;
	if (console1.backscroll > console1.buffer_rows - (clheight >> 3) - 1)
		console1.backscroll = console1.buffer_rows - (clheight >> 3) - 1;
	//johnfitz

	console1.cursor_column = 0;
	console1.cursor_row++;
	memset (&console1.text[(console1.cursor_row%console1.buffer_rows)*console1.buffer_columns], ' ', console1.buffer_columns);
}



/*
================
Con_Print

Handles cursor positioning, line wrapping, etc
All console printing must go through this in order to be logged to disk
If no console is visible, the notify window will pop up.
================
*/
static void Con_Print (const char *txt)
{
	int		y;
	int		c, l;
	static int	cr;
	int		mask;


	//console1.backscroll = 0; //johnfitz -- better console scrolling

	if (txt[0] == 1)
	{
		mask = 128;		// go to colored text
		S_LocalSound ("misc/talk.wav"); // play talk wav
		txt++;
	}
	else if (txt[0] == 2)
	{
		mask = 128;		// go to colored text
		txt++;
	}
	else
	{
		mask = 0;
	}

	while ( (c = *txt) )
	{
	// count word length
		for (l = 0 ; l < console1.buffer_columns ; l++)
			if ( txt[l] <= ' ')
				break;

	// word wrap
		if (l != console1.buffer_columns && (console1.cursor_column + l > console1.buffer_columns) )
			console1.cursor_column = 0;

		txt++;

		if (cr)
		{
			console1.cursor_row--;
			cr = false;
		}

		if (!console1.cursor_column)
		{
			Con_Linefeed ();
		// mark time for transparent overlay
			if (console1.cursor_row >= 0)
				console1.con_times[console1.cursor_row % (int)CONSOLE_NOTIFY_TIMES] = realtime;
		}

		switch (c)
		{
		case '\n':
			console1.cursor_column = 0;
			break;

		case '\r':
			console1.cursor_column = 0;
			cr = 1;
			break;

		default:	// display character and advance
			y = console1.cursor_row % console1.buffer_rows;
			console1.text[y*console1.buffer_columns+console1.cursor_column] = c | mask;
			console1.cursor_column++;
			if (console1.cursor_column >= console1.buffer_columns)
				console1.cursor_column = 0;
			break;
		}
	}
}


/*
================
Con_DebugLog
================
*/

void Con_DebugLog (const char *filename, const char *fmt, ...)
{
    FILE *f;
	VA_EXPAND (msg, 1024, fmt) 

	f = FS_fopen_write (filename, "a");
	fprintf (f, "%s", msg);
    FS_fclose(f);
}


/*
================
Con_Printf

Handles cursor positioning, line wrapping, etc
================
*/
#define	MAXPRINTMSG	4096
void Con_Printf (const char *fmt, ...)
{
	static cbool	inupdate;

	VA_EXPAND (msg, MAXPRINTMSG, fmt) 

// also echo to debugging console
	System_Printf ("%s", msg); // also echo to debugging console

// log all messages to file
	if (con_debuglog)
		Con_DebugLog (qpath_to_url("qconsole.log"), "%s", msg);

	if (!console1.initialized)
		return;

	if (isDedicated)
		return;		// no graphics mode

// write it to the scrollable buffer
	Con_Print (msg);

#if 1 // Engine X doesn't have ... eliminating this means black screen longer at startup :(

// update the screen if the console is displayed
	if (cls.signon != SIGNONS && !scr_disabled_for_loading )
	{
	// protect against infinite loop if something in SCR_UpdateScreen calls
	// Con_Printd
		if (!inupdate)
		{
			inupdate = true;
			SCR_UpdateScreen ();
			inupdate = false;
		}
	}
#endif
}


/*
================
Con_Warning -- johnfitz -- prints a warning to the console
================
*/
void Con_Warning (const char *fmt, ...)
{
	VA_EXPAND (msg, MAXPRINTMSG, fmt) 

	Con_SafePrintf ("%sWarning: ", isDedicated ? "": "\x02");
	Con_Printf ("%s", msg);
}

/*
================
Con_DPrintf

A Con_Printf that only shows up if the "developer" cvar is set
================
*/
void Con_DPrintf (const char *fmt, ...)
{
	// don't confuse non-developers with techie stuff...
	if (developer.value)
	{
		VA_EXPAND (msg, MAXPRINTMSG, fmt) 

		Con_SafePrintf ("%s", msg); //johnfitz -- was Con_Printf
	}
}

/*
================
Con_DPrintf2 -- johnfitz -- only prints if "developer" >= 2

currently not used
================
*/
void Con_DPrintf2 (const char *fmt, ...)
{

	if (developer.value >= 2)
	{
		VA_EXPAND (msg, MAXPRINTMSG, fmt)

		Con_Printf ("%s", msg);
	}
}


/*
==================
Con_SafePrintf

Okay to call even when the screen can't be updated
==================
*/
void Con_SafePrintf (const char *fmt, ...)
{
	int			temp;

	VA_EXPAND (msg, MAXPRINTMSG, fmt) 

	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true; // temped
	Con_Printf ("%s", msg);
	scr_disabled_for_loading = temp;
}

/*
================
Con_CenterPrintf -- johnfitz -- pad each line with spaces to make it appear centered
================
*/
void Con_CenterPrintf (int linewidth, const char *fmt, ...) __attribute__((__format__(__printf__,2,3)));
void Con_CenterPrintf (int linewidth, const char *fmt, ...)
{
//	char	msg[MAXPRINTMSG]; //the original message
	char	line[MAXPRINTMSG]; //one line from the message
	char	spaces[21]; //buffer for spaces
	char	*src, *dst;
	int		len, s;
	VA_EXPAND (msg, MAXPRINTMSG, fmt) 

	linewidth = c_min (linewidth, console1.buffer_columns);
	for (src = msg; *src; )
	{
		dst = line;
		while (*src && *src != '\n')
			*dst++ = *src++;
		*dst = 0;
		if (*src == '\n')
			src++;

		len = strlen(line);
		if (len < linewidth)
		{
			s = (linewidth-len)/2;
			memset (spaces, ' ', s);
			spaces[s] = 0;
			Con_Printf ("%s%s\n", spaces, line);
		}
		else
			Con_Printf ("%s\n", line);
	}
}

/*
==================
Con_LogCenterPrint -- johnfitz -- echo centerprint message to the console
==================
*/
void Con_LogCenterPrint (const char *str)
{

	if (!strcmp(str, cl.lastcenterstring))
		return; //ignore duplicates

	if (cl.gametype == GAME_DEATHMATCH && scr_logcenterprint.value != 2)
		return; //don't log in deathmatch

	c_strlcpy (cl.lastcenterstring, str);

	if (scr_logcenterprint.value)
	{
		Con_Printf ("%s", Con_Quakebar(40));
		Con_CenterPrintf (40, "%s\n", str);
		Con_Printf ("%s", Con_Quakebar(40));
		Con_ClearNotify ();
	}
}

/*
==============================================================================

	TAB COMPLETION

==============================================================================
*/

//johnfitz -- tab completion stuff
//unique defs
char key_tabpartial[CONSOLE_MAX_CMDLINE];
cbool key_inpartial = false;
typedef struct tab_s
{
	const char	*name;
	const char	*type;
	struct tab_s	*next;
	struct tab_s	*prev;
} tab_t;
tab_t	*tablist;

//defs from elsewhere

/*
============
AddToTabList -- johnfitz

tablist is a doubly-linked loop, alphabetized by name
============
*/
void AddToTabList (const char *name, const char *type)
{
	tab_t	*t,*insert;

	t = (tab_t *) Hunk_Alloc(sizeof(tab_t));
	t->name = name;
	t->type = type;

	if (!tablist) //create list
	{
		tablist = t;
		t->next = t;
		t->prev = t;
	}
	else if (strcmp(name, tablist->name) < 0) //insert at front
	{
		t->next = tablist;
		t->prev = tablist->prev;
		t->next->prev = t;
		t->prev->next = t;
		tablist = t;
	}
	else //insert later
	{
		insert = tablist;
		do
		{
			if (strcmp(name, insert->name) < 0)
				break;
			insert = insert->next;
		} while (insert != tablist);

		t->next = insert;
		t->prev = insert->prev;
		t->next->prev = t;
		t->prev->next = t;
	}
}

void BuildMediaList (clist_t* headnode, const char *tagstr, const char *partial)
{
	clist_t	*item;
	int				len = strlen(partial);

	tablist = NULL;

	for (item = headnode ; item ; item=item->next)
		if (!strncmp (partial,item->name, len))
			AddToTabList (item->name, "");
}

/*
============
BuildTabList -- johnfitz
============
*/
void BuildTabList (const char *partial)
{
	extern	cmd_alias_t		*cmd_alias;
	extern	cmd_function_t	*cmd_functions;

	cmd_alias_t		*alias;
	cvar_t			*cvar;
	cmd_function_t	*cmd;
	int				len;

	tablist = NULL;
	len = strlen(partial);

	cvar = Cvar_FindAfter ("", CVAR_NONE);
	for ( ; cvar ; cvar=cvar->next)
	{
		if (cvar->flags & CVAR_COURTESY) // Baker: These don't show
			continue;

		if (!strncmp (partial, cvar->name, len))
			AddToTabList (cvar->name, "cvar");
	}

	for (cmd = cmd_functions ; cmd ; cmd = cmd->next)
		if (!strncmp (partial,cmd->name, len))
			AddToTabList (cmd->name, "command");

	for (alias = cmd_alias ; alias ; alias = alias->next)
		if (!strncmp (partial, alias->name, len))
			AddToTabList (alias->name, "alias");
}



/*
============
Con_TabComplete -- Baker
============
*/

int in_match;
char *match_start;
char *match_end;
// Tab complete doesn't know when it is invalidated.
// Future?  In theory, any keypress and/or exit of the console should invalidate it.
#pragma message ("Baker: Tab Complete 3 problems")
#pragma message ("Baker: First, it is super-intuitive as is which is nice.  And it works")
#pragma message ("Baker: However, the code is very long which I don't like")
#pragma message ("Baker: And I believe it could be way shorter and easier to quickly understand")
#pragma message ("Baker: I also which it could auto-complete things without a first letter")
#pragma message ("Baker: But we append a space at the end for convenience which is nice")
#pragma message ("Baker: But it makes it nigh impossible to know if a TAB represents current autocomplete or next one")
#pragma message ("Baker: Ideally shouldn't a backspace during autocomplete back out the whole untyped portion??")
#pragma message ("Baker: Barring the above, couldn't we autocomplete from nothing if there is no text after the cursor when partial state = 0")
void Con_TabComplete (cbool force_zero_length_completion)
{
	char		partial[CONSOLE_MAX_CMDLINE];
	const char		*match;
	static char	*c;
	tab_t		*t;
	int			mark, i;

	static list_type_t	last_complete_type = list_type_none;
	list_type_t	complete_type = list_type_none;
	char *start_of_line = &key_lines[edit_line][1];
	char		cmdname[80] = {0};


//
// Baker --- all this looking around is to determine the command
//

	// find start of current command
	char *last_char_before_cursor = key_lines[edit_line] + key_linepos - 1; //start one space left of cursor

	if (last_char_before_cursor >= start_of_line)
	{
		cbool in_arg2 = false;
		char *effective_start_of_line, *ending;

		// &key_lines[edit_line][1] is start of current line
		char *ch = last_char_before_cursor;

		while ( *ch != ';' && ch >= &key_lines[edit_line][1] )
			ch--;
		ch++; //start 1 char after the separator we just found

		// Forward past any white space
		while (*ch && *ch <= 32 && ch != last_char_before_cursor)
			ch++;

		effective_start_of_line = ch;
		// Now go forward until we find a space or end of line
		while (*ch && *ch != ' ' && ch < last_char_before_cursor)
			ch++;
		ch--;
		ending = ch; // cursor is at a blank

		// Check for an arg2.  We should be on the last letter of the first command.  Add 2.
		// There must be a at least 3 more characters in the effective command line
		// And the next one must be a space.
		if (&ch[3] < last_char_before_cursor && ch[1] == 32)
		{
			// Baker: Find a white space char then a real one and then a whitespace char means
			// we are at least in arg2 and what we found before should be ignored.

			ch +=2; // Advance it 2.  We know we can do this.
			do
			{
				// Now skip whitespace.
				while (*ch && *ch <= 32 && ch < last_char_before_cursor)
					ch++;

				// If we hit the end we can't be in arg2
				if (ch >= last_char_before_cursor)
					break;

				// Ok, we didn't hit the end and are in non-whitespace.
				// Try to find whitespace.
				// If we do, we have an arg2

				while (*ch && *ch > 32 && ch < last_char_before_cursor)
					ch++;

				if (*ch == 32 && ch != last_char_before_cursor)
				{
					in_arg2 = true;
//					Con_DPrintf ("arg2 found\n");
				}
			} while (0); // End of check


		}

		strncpy (cmdname, effective_start_of_line, ending - effective_start_of_line + 1);

		if (key_inpartial)
			complete_type = last_complete_type;
		else
		if (ending[1] == ' ' && in_arg2 == false)
		{
			int i;

			for (i = 1; i < MAX_LIST_TYPES; i++)
			{
				const char *matcher_string = list_info[i].associated_commands;

				if (matcher_string == NULL || !COM_ListMatch (matcher_string, cmdname))
					continue;

				// Found
				complete_type = i;
				Con_DPrintf ("Complete type is %s with extension %s\n", list_info[complete_type].description, list_info[complete_type].extension);

				// Build list of results
				break;
			}
		}
		else if (ending[1] == ' '  && in_arg2 == true && !strcmp("game",cmdname)) // Baker: A bit lazy today ...
			complete_type = list_type_game_hud;
		else if (ending[1] == ' '  && in_arg2 == true && !strcmp("setmusic",cmdname)) // Baker: A bit lazy today ...
			complete_type = list_type_mp3;
	}
	Con_DPrintf ("complete type is %s with cmdname of '%s' \n", list_info[complete_type].description, cmdname);

	if (complete_type != last_complete_type)
		key_inpartial = false;; // The type of completion has changed.

	last_complete_type = complete_type;
	in_match = true;

// if editline is empty, return
	if (key_lines[edit_line][1] == 0)
		return;

// get partial string (space -> cursor)
	if (key_inpartial == false) //first time through, find new insert point. (Otherwise, use previous.)
	{
		//work back from cursor until you find a space, quote, semicolon, or prompt
		c = key_lines[edit_line] + key_linepos - 1; //start one space left of cursor
		while (*c!=' ' && *c!='\"' && *c!=';' && c!=key_lines[edit_line])
			c--;
		c++; //start 1 char after the separator we just found
	}
	for (i = 0; c + i < key_lines[edit_line] + key_linepos; i++)
		partial[i] = c[i];
	partial[i] = 0;
	// Record c and i
	match_start=c;
	match_end = &c[i-1];

	// In the middle of a word
	if (c[i] != ' ' && c[i])
		return;
//if partial is empty, return
	if (partial[0] == 0 && force_zero_length_completion == false)
		return;

//trim trailing space becuase it screws up string comparisons
	if (i > 0 && partial[i-1] == ' ')
		partial[i-1] = 0;

	match_end = match_start + strlen(partial) - 1;
// find a match
	mark = Hunk_LowMark();
	if (key_inpartial == false) //first time through
	{
		c_strlcpy (key_tabpartial, partial);
		key_inpartial = true;
		switch (complete_type)
		{
		case list_type_map:
		case list_type_game:
		case list_type_game_hud:
		case list_type_demo:
		case list_type_savegame:
		case list_type_key:
#ifdef GLQUAKE_TEXTUREMODES
		case list_type_texmode:
#endif // GLQUAKE_TEXTUREMODES
		case list_type_config:
		case list_type_sky:
		case list_type_mp3:
		case list_type_sound:
			BuildMediaList (list_info[complete_type].plist, list_info[complete_type].description, key_tabpartial);
			break;

		case list_type_none:
		default:
			BuildTabList (key_tabpartial);
		}

		if (!tablist)
		{
			Con_Printf ("No %s starting with %s%s\n", list_info[complete_type].description, key_tabpartial, (complete_type == list_type_map) ? " in current gamedir" : "");
			return;
		}

		//print list
		t = tablist;
		do
		{
			if (t->type[0])
				Con_SafePrintf("   %s (%s)\n", t->name, t->type);
			else Con_SafePrintf("   %s\n", t->name);
			t = t->next;
		} while (t != tablist);

		//get first match
		match = tablist->name;
	}
	else
	{
		switch (complete_type)
		{
		case list_type_map:
		case list_type_game:
		case list_type_game_hud:
		case list_type_demo:
		case list_type_savegame:
		case list_type_key:
		case list_type_config:
#ifdef GLQUAKE_TEXTUREMODES
		case list_type_texmode:
#endif // GLQUAKE_TEXTUREMODES
		case list_type_sky:
		case list_type_mp3:
		case list_type_sound:
			BuildMediaList (list_info[complete_type].plist, list_info[complete_type].description, key_tabpartial);
			break;

		case list_type_none:
		default:
			BuildTabList (key_tabpartial);
		}


		if (!tablist)
		{
			Con_Printf ("No %s starting with %s%s\n", list_info[complete_type].description, key_tabpartial, (complete_type == list_type_map) ? " in current gamedir" : "");
			return;
		}

		//find current match -- can't save a pointer because the list will be rebuilt each time
		t = tablist;
		do
		{
			if (!strcmp(t->name, partial))
			{
				break;
			}
			t = t->next;
		} while (t != tablist);

		//use prev or next to find next match
		match = Key_Shift_Down () ? t->prev->name : t->next->name;
	}
	Hunk_FreeToLowMark(mark); //it's okay to free it here because match is a pointer to persistent data

#if 1
// insert new match into edit line
	c_strlcpy (partial, match); //first copy match string
	strlcat (partial, key_lines[edit_line] + key_linepos, CONSOLE_MAX_CMDLINE); //then add chars after cursor
#pragma message ("Baker: I think this is safe, yet it really needs to use strlcpy with some size set")
	strcpy (c, partial); // now copy all of this into edit line
#else
	// Baker: This ends up being very, very hard to calculate
	// match start is the start of the bronzed
	// match end is the end of the word
	// Can't be done because we might be overwriting a stale autocomplete too.
	// The idea of inserting into the buffer here won't work
	// However, clipboard might be a different story
	c_strlcpy (partial, match); //first copy match string
	String_Edit_Insert_At (key_lines[edit_line], partial, key_linepos, sizeof(partial));


#endif
	key_linepos = c - key_lines[edit_line] + strlen(match); //set new cursor position
	key_sellength = 0;

// if cursor is at end of string, let's append a space to make life easier
	if (key_lines[edit_line][key_linepos] == 0)
	{
		key_lines[edit_line][key_linepos] = ' ';
		key_linepos++;
		key_lines[edit_line][key_linepos] = 0;
	}
}

/*
==============================================================================

DRAWING

==============================================================================
*/

/*
================
Con_DrawNotify

Draws the last few lines of output transparently over the game top
================
*/
void Con_DrawNotify (void)
{
	int	i, x, v;
	const char	*text;
	float	time;
	extern char chat_buffer[];

	Draw_SetCanvas (CANVAS_CONSOLE); //johnfitz
	v = vid.conheight; //johnfitz

	for (i = console1.cursor_row - scr_con_notifylines.value + 1 ; i <= console1.cursor_row ; i++)
	{
		if (i < 0)
			continue;
		time = console1.con_times[i % CONSOLE_NOTIFY_TIMES];
		if (time == 0)
			continue;
		time = realtime - time;
		if (time > scr_con_notifytime.value)
			continue;
		text = console1.text + (i % console1.buffer_rows)*console1.buffer_columns;


		for (x = 0 ; x < console1.buffer_columns ; x++)
			Draw_Character ( (x+1)<<3, v, text[x]);

		v += 8;

#ifdef GLQUAKE // Baker: Eliminate in future
		glquake_scr_tileclear_updates = 0; //johnfitz
#else
		winquake_scr_copytop = 1;
#endif // GLQUAKE vs. WINQUAKE
		clearnotify = 0; // Both
	}

	if (key_dest == key_message)
	{
		char *say_prompt; //johnfitz

		x = 0;

		//johnfitz -- distinguish say and say_team
		if (chat_team)
			say_prompt = "say_team:";
		else
			say_prompt = "say:";
		//johnfitz

		Draw_String (8, v, say_prompt); //johnfitz

		while(chat_buffer[x])
		{
			Draw_Character ( (x+strlen(say_prompt)+2)<<3, v, chat_buffer[x]); //johnfitz
			x++;
		}
		Draw_Character ( (x+strlen(say_prompt)+2)<<3, v, 10+((int)(realtime * CONSOLE_CURSOR_SPEED)&1)); //johnfitz
		v += 8;

#ifdef GLQUAKE // Baker: Eliminate in future
		glquake_scr_tileclear_updates = 0; //johnfitz
#else
		winquake_scr_copytop = 1;
#endif // GLQUAKE vs. WINQUAKE
		clearnotify = 0;
	}
}

/*
================
Con_DrawInput -- johnfitz -- modified to allow insert editing

The input line scrolls horizontally if typing goes beyond the right edge
================
*/
extern qpic_t *pic_ovr, *pic_ins; //johnfitz -- new cursor handling

void Con_DrawInput (void)
{
	int		i, len, ofs = 0;
	char	c[CONSOLE_MAX_CMDLINE], *text;

	if (key_dest != key_console && !console1.forcedup)
		return;		// don't draw anything

	c_strlcpy (c, key_lines[edit_line]);
	text = c;

// pad with one space so we can draw one past the string (in case the cursor is there)
	len = strlen(key_lines[edit_line]);
	text[len] = ' ';
	text[len+1] = 0;

// prestep if horizontally scrolling
	if (key_linepos >= console1.buffer_columns)
	{
		ofs = 1 + key_linepos - console1.buffer_columns;
		text += 1 + key_linepos - console1.buffer_columns;
	}

// draw input string
#pragma message ("Baker: I don't like this")
	// key_selstart key_sellength
	if (1 || key_sellength)
	{
//Draw_Pic ((key_linepos + 1 - ofs)<<3, vid.conheight - 16, key_insert ? pic_ins : pic_ovr);
		//(key_linepos + 1 - ofs)<<3; 
//		key_sellength=-1;
		{
		int startx = (key_sellength > 0 ? key_linepos - key_sellength: key_linepos);
		int endx = startx + abs(key_sellength) - 1;
		int netstartx = startx - ofs;
		int netendx = endx - ofs;
		int pixelstart = (netstartx + 1) * 8;
		int pixelend1 = pixelstart +  abs(key_sellength) * 8;
		int pixellen;
		// Clampage ...
		if (pixelstart < 8)
			pixelstart = 8;
		if (pixelend1 > vid.conwidth - 8)
			pixelend1 = vid.conwidth - 8;
		pixellen = pixelend1 - pixelstart;
		//		VID_Local_Set_Window_Caption (   va("%i %i --> ofs= %i -- startX endX %i %i ofs -- net %i %i pix: %i %i", key_linepos, key_sellength, ofs, startx, endx, netstartx, netendx, pixelstart, pixellen));
		if (key_sellength )
			Draw_Fill (pixelstart, vid.conheight - 16, pixellen + 1, 9, 254, 1);
		}
	}

	for (i = 0; i <= (int)(strlen(text) - 1) && i < console1.buffer_columns; i++) //only write enough letters to go from *text to cursor
	{
		const char *xmatch_start = (match_start - key_lines[edit_line]) + c;
		const char *xmatch_end = xmatch_start + strlen(key_tabpartial) - 1;
//		char *last_char_before_cursor = key_lines[edit_line] + key_linepos - 1;
		if (key_tabpartial[0] && xmatch_start <= &text[i] &&  &text[i] <= xmatch_end)
		{
			Draw_Character ((i+1)<<3, vid.conheight - 16, text[i] | 128); // Bronze tab completion in progress
		}
		else Draw_Character ((i+1)<<3, vid.conheight - 16, text[i]);

	}

// johnfitz -- new cursor handling
	if (!((int)((realtime-key_blinktime) * CONSOLE_CURSOR_SPEED) & 1))
#ifdef GLQUAKE // Baker: Eliminate in future
#if 0
		Draw_Pic ((key_linepos + 1 - ofs)<<3, vid.conheight - 16, key_insert ? pic_ins : pic_ovr);
#else
		// Draw the length doing some sort of XOR draw?
		// What we really want are the characters to be XOR drawn with lite background, dark foreground
		// How can we do this?
		Draw_Pic ((key_linepos + 1 - ofs)<<3, vid.conheight - 16, key_insert ? pic_ins : pic_ovr);
#endif
#else
		if (console1.visible_lines > 16)
			Draw_Fill ((key_linepos+1 - ofs) << 3, vid.conheight - 16, key_insert ? 2 : 8, 8, 254, 1);
#endif // GLQUAKE vs. WINQUAKE
}

/*
================
Con_DrawConsole -- johnfitz -- heavy revision

Draws the console with the solid background
The typing input line at the bottom should only be drawn if typing is allowed
================
*/
void Con_DrawConsole (float pct, cbool drawinput)
{
	int				i, j, x, y, rows;
	int				sb;

	const char	*text;
	char			ver[32];

	if (pct <= 0)
		return;

// draw the background
	Draw_SetCanvas (CANVAS_CONSOLE);
	Draw_ConsoleBackground ();


	// Baker: Don't draw the console text if menu is being drawn.
	if (key_dest == key_menu && m_state)
		return;

// draw the text
	rows = (console1.visible_lines_conscale + 7)/8;
	y = vid.conheight - rows * 8;
	rows -= 2; //for input and version lines
	sb = (console1.backscroll) ? 2 : 0;

	for (i = console1.cursor_row - rows + 1 ; i <= console1.cursor_row - sb ; i++, y+=8)
	{
		j = i - console1.backscroll;
		if (j<0)
			j = 0;
		text = console1.text + (j % console1.buffer_rows) * console1.buffer_columns;

		for (x=0 ; x < console1.buffer_columns ; x++)
			Draw_Character ( (x+1)<<3, y, text[x]);
	}

// draw scrollback arrows
	if (console1.backscroll)
	{
		y+=8; // blank line
		for (x=0 ; x < console1.buffer_columns ; x+=4)
			Draw_Character ((x+1)<<3, y, '^');
		y+=8;
	}

// draw the input prompt, user text, and cursor
	if (drawinput)
		Con_DrawInput ();

//draw version number in bottom right
	y+=8;
	c_snprintf2 (ver, "%s %1.2f"/*" beta2"*/, ENGINE_NAME, (float)ENGINE_VERSION);
	for (x=0; x < (int) strlen(ver); x++)
		Draw_Character ((console1.buffer_columns-strlen(ver)+x+2)<<3, y, ver[x] /*+ 128*/);
}


/*
==================
Con_NotifyBox
==================
*/
void Con_NotifyBox (const char *text)
{
	double		t1, t2;

// during startup for sound / cd warnings
	Con_Printf ("\n\n%s", Con_Quakebar(40)); //johnfitz
	Con_Printf ("%s", text);
	Con_Printf ("Press a key.\n");
	Con_Printf ("%s", Con_Quakebar(40)); //johnfitz

	key_count = -2; // wait for a key down and up
	Key_SetDest (key_console);

	do
	{
		t1 = System_DoubleTime ();
		SCR_UpdateScreen ();
		System_SendKeyEvents ();
		t2 = System_DoubleTime ();
		realtime += t2 - t1; // make the cursor blink
	} while (key_count < 0);

	Con_Printf ("\n");
	Key_SetDest (key_game);
	realtime = 0; // put the cursor back to invisible
}

